@@ -23,11 +23,8 @@
 #include "buffer.h"
 
 typedef struct Buffer_t {
-	Class_t *Class;
-	int refs;
-	Stream_t *Next;
-	Stream_t *Buffer;
-	
+	struct Stream_t head;
+
 	size_t size;     	/* size of read/write buffer */
 	int dirty;	       	/* is the buffer dirty? */
 
@@ -69,7 +66,11 @@ static int _buf_flush(Buffer_t *Buffer)
 {
 	ssize_t ret;
 
-	if (!Buffer->Next || !Buffer->dirty)
+#ifdef HAVE_ASSERT_H
+	assert(Buffer->head.Next != NULL);
+#endif
+	
+	if (!Buffer->dirty)
 		return 0;
 #ifdef DEBUG
 	fprintf(stderr, "write %08x -- %02x %08x %08x\n",
@@ -79,15 +80,17 @@ static int _buf_flush(Buffer_t *Buffer)
 		Buffer->dirty_end - Buffer->dirty_pos);
 #endif
 
-	ret = force_write(Buffer->Next,
-			  Buffer->buf + Buffer->dirty_pos,
-			  Buffer->current + (mt_off_t) Buffer->dirty_pos,
-			  Buffer->dirty_end - Buffer->dirty_pos);
-	if(ret != (signed int) (Buffer->dirty_end - Buffer->dirty_pos)) {
-		if(ret < 0)
-			perror("buffer_flush: write");
-		else
-			fprintf(stderr,"buffer_flush: short write\n");
+	ret = force_pwrite(Buffer->head.Next,
+			   Buffer->buf + Buffer->dirty_pos,
+			   Buffer->current + (mt_off_t) Buffer->dirty_pos,
+			   Buffer->dirty_end - Buffer->dirty_pos);
+	if(ret < 0) {
+		perror("buffer_flush: write");
+		return -1;
+	}
+	
+	if((size_t) ret != Buffer->dirty_end - Buffer->dirty_pos) {
+		fprintf(stderr,"buffer_flush: short write\n");
 		return -1;
 	}
 	Buffer->dirty = 0;
@@ -146,16 +149,17 @@ static position_t isInBuffer(Buffer_t *This, mt_off_t start, size_t *len)
 	}
 }
 
-static ssize_t buf_read(Stream_t *Stream, char *buf, mt_off_t start, size_t len)
+static ssize_t buf_pread(Stream_t *Stream, char *buf,
+			 mt_off_t start, size_t len)
 {
 	size_t length;
 	size_t offset;
 	char *disk_ptr;
 	ssize_t ret;
-	DeclareThis(Buffer_t);	
+	DeclareThis(Buffer_t);
 
 	if(!len)
-		return 0;	
+		return 0;
 
 	/*fprintf(stderr, "buf read %x   %x %x\n", Stream, start, len);*/
 	switch(isInBuffer(This, start, &len)) {
@@ -166,10 +170,10 @@ static ssize_t buf_read(Stream_t *Stream, char *buf, mt_off_t start, size_t len)
 			maximize(length, This->size - This->cur_size);
 
 			/* read it! */
-			ret=READS(This->Next,
-				  This->buf + This->cur_size,
-				  This->current + (mt_off_t) This->cur_size,
-				  length);
+			ret=PREADS(This->head.Next,
+				   This->buf + This->cur_size,
+				   This->current + (mt_off_t) This->cur_size,
+				   length);
 			if ( ret < 0 )
 				return ret;
 			This->cur_size += (size_t) ret;
@@ -192,11 +196,11 @@ static ssize_t buf_read(Stream_t *Stream, char *buf, mt_off_t start, size_t len)
 	return (ssize_t) len;
 }
 
-static ssize_t buf_write(Stream_t *Stream, char *buf,
-			 mt_off_t start, size_t len)
+static ssize_t buf_pwrite(Stream_t *Stream, char *buf,
+			  mt_off_t start, size_t len)
 {
 	char *disk_ptr;
-	DeclareThis(Buffer_t);	
+	DeclareThis(Buffer_t);
 	size_t offset=0;
 
 	if(!len)
@@ -223,11 +227,12 @@ static ssize_t buf_write(Stream_t *Stream, char *buf,
 				size_t readSize;
 				ssize_t ret;
 				size_t bytes_read;
-				
+
 				readSize = This->cylinderSize -
 					(size_t)(This->current % (mt_off_t) This->cylinderSize);
 
-				ret=READS(This->Next, This->buf, (mt_off_t)This->current, readSize);
+				ret=PREADS(This->head.Next, This->buf,
+					   (mt_off_t)This->current, readSize);
 				/* read it! */
 				if ( ret < 0 )
 					return ret;
@@ -258,8 +263,8 @@ static ssize_t buf_write(Stream_t *Stream, char *buf,
 			offset = OFFSET;
 			maximize(len, This->size - offset);
 			This->cur_size += len;
-			if(This->Next->Class->pre_allocate)
-				PRE_ALLOCATE(This->Next, cur_end(This));
+			if(This->head.Next->Class->pre_allocate)
+				PRE_ALLOCATE(This->head.Next, cur_end(This));
 			break;
 		case INSIDE:
 			/* nothing to do */
@@ -291,7 +296,7 @@ static ssize_t buf_write(Stream_t *Stream, char *buf,
 		This->dirty_pos = ROUND_DOWN(offset, This->sectorSize);
 	if(!This->dirty || offset + len > This->dirty_end)
 		This->dirty_end = ROUND_UP(offset + len, This->sectorSize);
-	
+
 	if(This->dirty_end > This->cur_size) {
 		fprintf(stderr,
 			"Internal error, dirty end too big dirty_end=%x cur_size=%x len=%x offset=%d sectorSize=%x\n",
@@ -337,8 +342,10 @@ static int buf_free(Stream_t *Stream)
 }
 
 static Class_t BufferClass = {
-	buf_read,
-	buf_write,
+	0,
+	0,
+	buf_pread,
+	buf_pwrite,
 	buf_flush,
 	buf_free,
 	0, /* set_geom */
@@ -353,12 +360,12 @@ Stream_t *buf_init(Stream_t *Next, size_t size,
 		   size_t sectorSize)
 {
 	Buffer_t *Buffer;
-	Stream_t *Stream;
 
 #ifdef HAVE_ASSERT_H
 	assert(size != 0);
 	assert(cylinderSize != 0);
 	assert(sectorSize != 0);
+	assert(Next != NULL);
 #endif
 
 	if(size % cylinderSize != 0) {
@@ -370,19 +377,13 @@ Stream_t *buf_init(Stream_t *Next, size_t size,
 		exit(1);
 	}
 
-	if(Next->Buffer){
-		Next->refs--;
-		Next->Buffer->refs++;
-		return Next->Buffer;
-	}
-
-	Stream = (Stream_t *) malloc (sizeof(Buffer_t));
-	if(!Stream)
+	Buffer = New(Buffer_t);
+	if(!Buffer)
 		return 0;
-	Buffer = (Buffer_t *) Stream;
+	init_head(&Buffer->head, &BufferClass, Next);
 	Buffer->buf = malloc(size);
 	if ( !Buffer->buf){
-		Free(Stream);
+		Free(Buffer);
 		return 0;
 	}
 	Buffer->size = size;
@@ -396,11 +397,6 @@ Stream_t *buf_init(Stream_t *Next, size_t size,
 	Buffer->current = 0L;
 	Buffer->cur_size = 0; /* buffer currently empty */
 
-	Buffer->Next = Next;
-	Buffer->Class = &BufferClass;
-	Buffer->refs = 1;
-	Buffer->Buffer = 0;
-	Buffer->Next->Buffer = (Stream_t *) Buffer;
-	return Stream;
+	return &Buffer->head;
 }
 
