@@ -18,10 +18,10 @@
 #include "sysincludes.h"
 #include "msdos.h"
 
-const char *mversion="4.0.33";
+const char *mversion="4.0.38";
 
 /* Multiple releases on same day should be marked with (b), (cd), (d) after
  * date string below */
-const char *mdate = "July 17th, 2021";
+const char *mdate = "March 3rd, 2022";
 
-const char *mformat_banner = "MTOO4033";
+const char *mformat_banner = "MTOO4038";
